///////////////////////////////////////////////////////////////////////////////////////////////////
////MystyMood multitexture-terrainshader that makes it easy to implement terrains made 
////with EarthSculptor (http://www.earthsculptor.com/) into 3dgs. It's able to blend 4 textures and 
////I also included an underwater multicaustics wave sun reflectionand and easy shadowmap generation.
////Contains ffe fallbacks, supports dynamic lights and sunlight and A6 stencil shadows
////
////NEEDS: terrain_materials.wdl
////
////Jun 07 by www.loopix-project.com
///////////////////////////////////////////////////////////////////////////////////////////////////

define		terr_add_red,skill1; //add a color tint
define		terr_add_green,skill2;
define		terr_add_blue,skill3;
define		terr_fogcol_fac,skill4;//factor of focgcolor1 influence
define		terr_caust_speed, skill5;//speed of the underwater caustics (sun-wave reflection)
define		terr_caust_scale, skill6;//scale of the underwater caustics
define		terr_caust_strenght, skill7;

define		terr_id_number, skill8;//scale of the underwater caustics

define		terr_ffe_only,flag1;//check this flag if you don't want shader terrain (looks not so good)
define		terr_shad_gen,flag2;//check this flag for generating the shadowmap (static-shadowmap...uncheck for normal terrain use)
define		terr_no_caustics,flag3;//flag to disable underwater caustics

////this loads the terrain_material.wdl wich is needed
include <terrain_materials.wdl>;

//string terr_shadowmap_string = "terr_shadowmap_1.tga";
string terr_shadowmap_str = "terr_shadowmap_";
string terr_shadowmap_tempstr1;
string terr_shadowmap_tempstr2;

string terr_shadowmap_filetype = ".tga";//for terr_shadowmap: terr_shadowmap.tga...

////SHADOWMAP GENERATION ADJUSTMENTS TO BE MADE HERE! 
var shadow_brightness=120;//brightness of shadows...(0=black, 255=white)
var shadow_pixel_z_adjust = -1;//play with this within a range of -5 to 5! This affects the shadow offset from the object
var blur_passes = 1;//the higher the softer the borders of the shadows


bmap* canvas_shadowmap;
var canvas_shadowmap_size[2];

font fnt_arial = "arial",0,25; 

string  strg_wait_shadow_gen = "Please wait...shadowmap-generation is progressing...\n This can take some time...";
string  strg_shadow_finished = "Shadowmap is now generated and saved to your project folder!";
string  strg_press_shadow_gen = "Press 'G' to generate shadowmap!";


text txt_wait_shadow_gen
{
	font = fnt_arial;
	layer = 10;
	flags = center_x,center_y,outline;
	string = strg_wait_shadow_gen;
	red = 255;
	green = 0;
	blue = 0;
}

text txt_shadow_finished
{
	font = fnt_arial;
	layer = 10;
	flags = center_x,center_y,outline;
	string = strg_shadow_finished;
	red = 0;
	green = 255;
	blue = 0;
}

text txt_press_shadow_gen
{
	font = fnt_arial;
	layer = 10;
	flags = center_x,center_y,outline;
	string = strg_press_shadow_gen;
	red = 0;
	green = 0;
	blue = 255;
}

var_nsave shadow_fhandle;
//---------------------------------------------------------write tga
function shadow_write8(shadow_byte) // write char
{
	file_asc_write(shadow_fhandle,shadow_byte);
}

function shadow_write16(shadow_short) // write unsigned shadow_short
{
	file_asc_write(shadow_fhandle,shadow_short&255);
	file_asc_write(shadow_fhandle,(shadow_short>>8)&255);
}

function write_shadowmap_tga(shadowmap_file_name)
{
	var i;
	var format;
	var pixel;
	var pixelalpha;
	var canvas_shadowmap_size[2];

	canvas_shadowmap_size.x=bmap_width(canvas_shadowmap);
	canvas_shadowmap_size.y=bmap_height(canvas_shadowmap);
	format=bmap_lock(canvas_shadowmap,0);

	shadow_fhandle=file_open_write(shadowmap_file_name);
	//------------------------------------write header
	shadow_write8(0);
	shadow_write8(0);
	shadow_write8(2); // image type // 10 = RLE compressed tga, 2 = uncompressed tga
	shadow_write16(0);
	shadow_write16(0);
	shadow_write8(0);
	shadow_write16(0);
	shadow_write16(0);
	shadow_write16(canvas_shadowmap_size.x); // width
	shadow_write16(canvas_shadowmap_size.y); // height
	shadow_write8(24);//or 32 for alphachanel
	shadow_write8(0);
	//------------------------------------write image data
	i=0;
	while(i<canvas_shadowmap_size.x*canvas_shadowmap_size.y)
	{
		pixel = pixel_for_bmap(canvas_shadowmap,i%canvas_shadowmap_size.x,(canvas_shadowmap_size.y-1)-int(i/canvas_shadowmap_size.x));
		pixel_to_vec(temp,pixelalpha,format,pixel);
		shadow_write8(temp.x); // b
		shadow_write8(temp.y); // g
		shadow_write8(temp.z); // r
		i+=1;
		if(i%max_loops == 0){wait(1);}
	}
	
	wait(1);
	
	file_close(shadow_fhandle);
	bmap_unlock(canvas_shadowmap);
	
	txt_wait_shadow_gen.visible = off;
	txt_shadow_finished.pos_x = screen_size.x/2;
	txt_shadow_finished.pos_y = screen_size.y/2;
	txt_shadow_finished.visible = on;
	wait(-4);
	txt_shadow_finished.visible = off;;
}

function shadow_get_xyz(shadow_px,shadow_py)
{
	var pixel_size[2];
	pixel_size.x=(my.max_x-my.min_x)/canvas_shadowmap_size.x;
	pixel_size.y=(my.max_y-my.min_y)/canvas_shadowmap_size.y;
	temp.x=((my.x-my.min_x)-(pixel_size.x*shadow_px))-pixel_size.x/2;
	temp.y=((my.y-my.min_y)-(pixel_size.y*shadow_py))-pixel_size.y/2;
	c_trace(vector(temp.x,temp.y,my.z+50000),vector(temp.x,temp.y,my.z-50000),ignore_models+ignore_sprites+ignore_maps);
	temp.z=target.z;
}

////blur
function blur(passes)
{
	var i;
	var px;
	var py;
	var format;
	var pixel;
	var pixelcolor[3];
	var pixelalpha;
	var sample1[3];
	var sample2[3];
	var sample3[3];
	var sample4[3];
	var sample5[3];
	var sample6[3];
	var sample7[3];
	var sample8[3];
	var sample9[3];

	i=0;

	while(i<passes)
	{
		py=0;
		while(py<canvas_shadowmap_size.y)
		{
			px=0;
			while(px<canvas_shadowmap_size.x)
			{
				format=bmap_lock(canvas_shadowmap,0);
				
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px,0,canvas_shadowmap_size.x-1),clamp(py,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample1,pixelalpha,format,pixel);
				
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px-1,0,canvas_shadowmap_size.x-1),clamp(py,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample2,pixelalpha,format,pixel);
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px-1,0,canvas_shadowmap_size.x-1),clamp(py+1,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample3,pixelalpha,format,pixel);
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px,0,canvas_shadowmap_size.x-1),clamp(py+1,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample4,pixelalpha,format,pixel);
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px+1,0,canvas_shadowmap_size.x-1),clamp(py+1,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample5,pixelalpha,format,pixel);
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px+1,0,canvas_shadowmap_size.x-1),clamp(py,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample6,pixelalpha,format,pixel);
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px+1,0,canvas_shadowmap_size.x-1),clamp(py-1,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample7,pixelalpha,format,pixel);
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px,0,canvas_shadowmap_size.x-1),clamp(py-1,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample8,pixelalpha,format,pixel);
				pixel=pixel_for_bmap(canvas_shadowmap,clamp((canvas_shadowmap_size.x-1)-px-1,0,canvas_shadowmap_size.x-1),clamp(py-1,0,canvas_shadowmap_size.x-1));
				pixel_to_vec(sample9,pixelalpha,format,pixel);
				
				pixelcolor.x=int((sample1.x*7+sample2.x*2+sample3.x+sample4.x*2+sample5.x+sample6.x*2+sample7.x+sample8.x*2+sample9.x)/19);
				pixelcolor.y=int((sample1.y*7+sample2.y*2+sample3.y+sample4.y*2+sample5.y+sample6.y*2+sample7.y+sample8.y*2+sample9.y)/19);
				pixelcolor.z=int((sample1.z*7+sample2.z*2+sample3.z+sample4.z*2+sample5.z+sample6.z*2+sample7.z+sample8.z*2+sample9.z)/19);
				
				pixel=pixel_for_vec(pixelcolor,100,format);
				pixel_to_bmap(canvas_shadowmap,(canvas_shadowmap_size.x-1)-px,py,pixel);
				
				bmap_unlock(canvas_shadowmap);
				
				px+=1;
			}
			py+=1;
			wait(1); // without a wait after each line the loop could get too big if the shadow map is huge
		}
		i+=1;
	}
	
	write_shadowmap_tga(terr_shadowmap_tempstr1);
}


function shadowmap_gen()
{
	//NOTE: YOUR TERRAIN ACTION MUST INCLUDE "C_SETMINMAX(ME); MY.POLYGON=ON;" AT THE VERY BEGINING!!!
	
	txt_press_shadow_gen.visible = off;
	
	var shadow_px;
	var shadow_py;
	var format;
	var pixel;
	var dist;
	
	canvas_shadowmap=bmap_for_entity(my,2); // the second texture of the terrain is the shadowmap
	canvas_shadowmap_size.x=bmap_width(canvas_shadowmap);
	canvas_shadowmap_size.y=bmap_height(canvas_shadowmap);

	you = my;//need this pointer to avoid terrain self-shadows

	shadow_py=0;
	while(shadow_py<canvas_shadowmap_size.y)
	{
		shadow_px=0;
		while(shadow_px<canvas_shadowmap_size.x)
		{
			shadow_get_xyz(shadow_px,shadow_py); // get the world coordinates of the pixel
			c_trace(sun_pos.x,vector(temp.x,temp.y,temp.z+(shadow_pixel_z_adjust)),ignore_passents); // trace from the sun to the pixel
			if(trace_hit && you != my) // draw shadow pixel if there is no obstacle. Only draws shadows if the you entity is not the ent_terrain
			{
				format=bmap_lock(canvas_shadowmap,0);
				pixel=pixel_for_vec(vector(shadow_brightness,shadow_brightness,shadow_brightness),100,format);
				pixel_to_bmap(canvas_shadowmap,(canvas_shadowmap_size.x-1)-shadow_px,shadow_py,pixel);
				bmap_unlock(canvas_shadowmap);
			}
			
			shadow_px+=1;
		}
		shadow_py+=1;
		
		txt_wait_shadow_gen.pos_x = screen_size.x/2;
		txt_wait_shadow_gen.pos_y = screen_size.y/2;
		txt_wait_shadow_gen.visible = on;
		
		wait(1); // without a wait after each line the loop could get too big if the shadow map is huge
	}

	blur(blur_passes); // do blur pass	
}


material mat_terr_shadowmap
{
	effect=
	"
	texture mtlSkin1;
	texture entSkin2;

	technique one_pass_shadow
	{
		pass p0
		{
			Texture[0]=<mtlSkin1>;
			Texture[1]=<entSkin2>;
			
			ColorArg1[0]=Texture;
			ColorOp[0]=Modulate2x;
			ColorArg2[0]=Diffuse;
			TexCoordIndex[0]=0;
			TextureTransformFlags[0]=Count2;
			TextureTransform[0]={1.0,0.0,0.0,0.0, // color map u scale
				0.0,1.0,0.0,0.0, // color map v scale
				0.0,0.0,0.0,0.0,
			0.0,0.0,0.0,0.0};
			
			ColorArg1[1]=Texture;
			ColorOp[1]=Modulate;
			ColorArg2[1]=Current;
			TexCoordIndex[1]=0;
			TextureTransformFlags[1]=Count2;
			TextureTransform[1]={1.0,0.0,0.0,0.0,
				0.0,1.0,0.0,0.0,
				0.0,0.0,0.0,0.0,
			0.0,0.0,0.0,0.0};
		}
	}
	";
}

//title:MystyMood Terrain Template
//image:loopix_logo.pcx
//action:act_terrain 
//skill1:terr_add_red 150
//help: terrain tint red 0-255
//skill2:terr_add_green 150
//help: terrain tint green 0-255
//skill3:terr_add_blue 150
//help: terrain tint blue 0-255
// section:
//skill4:terr_fogcol_fac 0.500
//help: influence of fogcolor1 on terrain 0...1
// section:
//skill5:terr_caust_speed 0.050
//help: underwater caustics speed
//skill6:terr_caust_scale 40.000
//help: underwater caustics scale
//section:
//skill7:terr_caust_strenght 150
//help: defines the visibility strenght of underwater caustics
//section:
//skill8:terr_id_number 1
//help: defines the terrain wich is needed to choose the corresponding material
// section:
//text: flag for non shader terrain only
//flag1:terr_ffe_only 0
// section:
//text: flag for shadow generation mode (pressing "g")...uncheck when done
//flag2:terr_shad_gen 0
// section:
//text: flag for disableing underwater caustics
//flag3:terr_no_caustics 0
//uses:terr_add_red,terr_add_green,terr_add_blue,terr_fogcol_fac,terr_ffe_only,terr_spec_pow_ffe,terr_shad_gen,terr_caust_speed,terr_caust_scale,terr_no_caustics,terr_caust_strenght,terr_id_number
action act_terrain {
	
	ent_terrain = me;
	
	//	my.x = 0;
	//	my.y = 0;
	//	my.z = 0;
	
	c_setminmax(me);
	wait(1);
	my.polygon = on;
	
	//if youn want to implement a SEED_IT (www.gameus.de) object seed-action, do it right here!
	
	
	//set defaults	
	if(!my.terr_fogcol_fac){my.terr_fogcol_fac = 0.2;}
	if(!my.terr_add_blue){my.terr_add_blue = 150;}
	if(!my.terr_add_green){my.terr_add_green = 150;}
	if(!my.terr_add_red){my.terr_add_red = 150;}
	if(!my.terr_caust_speed){my.terr_caust_speed = 0.05;}
	if(!my.terr_caust_scale){my.terr_caust_scale = 40;}
	if(!my.terr_caust_strenght){my.terr_caust_strenght = 150;}
	if(!my.terr_id_number){my.terr_id_number = 1;}
	
	////this is to find the according shadowmap
	str_cpy(terr_shadowmap_tempstr1,terr_shadowmap_str);
	str_for_num(terr_shadowmap_tempstr2,ent_terrain.terr_id_number);
	str_cat(terr_shadowmap_tempstr1,terr_shadowmap_tempstr2);
	str_cat(terr_shadowmap_tempstr1,terr_shadowmap_filetype);
	
	wait(1);
	
	var count;
	var a ;
	
	if(my.terr_shad_gen == off){
		
		if(d3d_shaderversion >= 2020 && my.terr_ffe_only == off){
			
			if(my.terr_id_number == 1){
				my.material = mat_terr_hlsl_1;
			}
//			if(my.terr_id_number == 2){
//				my.material = mat_terr_hlsl_2;
//			}
//			if(my.terr_id_number == 3){
//				my.material = mat_terr_hlsl_3;
//			}
//			if(my.terr_id_number == 4){
//				my.material = mat_terr_hlsl_4;
//			}
			////add more terrain materials the same way...
			
			my.skill44 = float(my.terr_caust_scale);	
			
			while(1){	
				my.skill41 = float(count);
				my.skill42 = float(-count);
				a += my.terr_caust_speed*time_step;
				count = a;		
				wait(1);	
			}	
		}
		else{
			
			if(my.terr_id_number == 1){
				my.material = mat_terr_ffe_1;
			}
//			if(my.terr_id_number == 2){
//				my.material = mat_terr_ffe_2;
//			}
//			if(my.terr_id_number == 3){
//				my.material = mat_terr_ffe_3;
//			}
//			if(my.terr_id_number == 4){
//				my.material = mat_terr_ffe_4;
//			}
			////add more terrain materials the same way...
		}	
	}
	else{//when in shadowmap generation mode...
		
		dynamic_day_night = 0;
//		sun_angle.pan = sun_azimuth;
//		sun_angle.tilt = sun_elevation;
		
		my.material = mat_terr_shadowmap;
		
		txt_press_shadow_gen.pos_x = screen_size.x/2;
		txt_press_shadow_gen.pos_y = screen_size.y/2;
		
		txt_press_shadow_gen.visible = on;
		
		while(key_g == 0){wait(1);}
		
		shadowmap_gen();
		
		txt_press_shadow_gen.visible = off;	
	}
}